#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "blackScholes.h"
#include "control3.h"
#include "util.h"

/* The exact value of control 1 used for the american estimate at time, path */
double innerControl1Exact(meshData *md, int time, int path){
  int max;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]);

  return(exp(-time*md->option.riskFreeRate)*euroCall(md->option.pData.price[time][path][max],
						     md->option.strikePrice,
						     md->option.riskFreeRate,
						     sqrt(md->option.pData.covariance[max][max]),
						     1.0,
						     md->option.pData.dividendRate[max]));
}


double innerControl2Exact(meshData *md, int time, int path){
  int max;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]);

  return(exp(md->option.pData.growthRate[max]-md->option.pData.dividendRate[max])*md->option.pData.price[time][path][max]);
}

double innerControl2Estimate(meshData *md, int time, int path){
  int max, toPath;
  double estimate, sumWeight;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]);

  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath < md->meshPaths; toPath++){
    sumWeight += md->weights[time][path][toPath];
    estimate += md->weights[time][path][toPath]*md->option.pData.price[time+1][toPath][max];
  }
  estimate /= sumWeight;
  return(estimate);
}

/* The mesh estimate of control1 at time, path */
double innerControl1Estimate(meshData *md, int time, int path){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r;
  
  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]); 

  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    sumWeight += md->weights[time][path][toPath];
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*(md->option.pData.price[time+1][toPath][max] - md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath]*deflatedPrice;
    /*    printf("E:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],deflatedPrice); */
  }
  estimate /= sumWeight;
  /*  printf("E:%lf\t%lf\n",sumWeight,exp(-time*md->option.riskFreeRate)*estimate);*/
  return(exp(-time*md->option.riskFreeRate)*estimate);
}

double innerControl2Variance(meshData *md, int time, int path){
  int max, toPath;
  double estimate,sumWeight;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]); 

  estimate = 0;
  sumWeight =0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    estimate += md->weights[time][path][toPath]*md->option.pData.price[time][path][max]*md->option.pData.price[time][path][max];
    sumWeight += md->weights[time][path][toPath];
    /*    printf("V:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],deflatedPrice);*/
  }
  estimate /= sumWeight;
  /*  printf("V:%lf\t%lf\n",sumWeight,estimate);*/
  return(exp(-2*time*md->option.riskFreeRate)*estimate);
}


/* The variance of the mesh estimate of control1 at time,path */
double innerControl1Variance(meshData *md, int time, int path){
  int max, toPath;
  double estimate,sumWeight,deflatedPrice;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]); 

  estimate = 0;
  sumWeight =0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate) * (md->option.pData.price[time+1][toPath][max]-md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath] *deflatedPrice*deflatedPrice;
    sumWeight += md->weights[time][path][toPath];
    /*    printf("V:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],deflatedPrice);*/
  }
  estimate /= sumWeight;
  /*  printf("V:%lf\t%lf\n",sumWeight,estimate);*/
  return(exp(-2*time*md->option.riskFreeRate)*estimate);
}

double innerControl2Correlation(meshData *md, int time, int path){
  int max, toPath;
  double estimate,sumWeight;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]); 

  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    estimate += md->weights[time][path][toPath] *md->option.pData.price[time+1][toPath][max]*md->newValue[time+1][toPath];
    sumWeight += md->weights[time][path][toPath];
    /*    printf("C:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],exp(-time*md->option.riskFreeRate)*deflatedPrice*md->newValue[time+1][toPath]);*/
  }
  estimate /= sumWeight;
  /*  printf("C:%lf\t%lf\n",sumWeight,estimate);*/
  return(estimate);
} 


/* The correlation of the american estimate and control1 */
double innerControl1Correlation(meshData *md, int time, int path){
int max, toPath;
  double estimate,sumWeight,deflatedPrice;

  max = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]); 

  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if(md->option.pData.price[time+1][toPath][max] > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate) * (md->option.pData.price[time+1][toPath][max]-md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath] * exp(-time*md->option.riskFreeRate)*deflatedPrice*md->newValue[time+1][toPath];
    sumWeight += md->weights[time][path][toPath];
    /*    printf("C:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],exp(-time*md->option.riskFreeRate)*deflatedPrice*md->newValue[time+1][toPath]);*/
  }
  estimate /= sumWeight;
  /*  printf("C:%lf\t%lf\n",sumWeight,estimate);*/
  return(estimate);
} 

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Control 3 estimators and exact calculations */

/* The following computes the exact value of control 3 using equations from
   Stulz. See control3.c */

double innerControl3Exact(meshData *md, int time, int path){
  int max1, max2, i;
  double maxval1, maxval2;

  /* At the given time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]);
  maxval1 = md->option.pData.price[time][path][max1];

  for(i=0; i<md->option.pData.underliers;i++){
    if(i != max1 && md->option.pData.price[time][path][i] > maxval2){
      max2 = i;
      maxval2 = md->option.pData.price[time][path][i];
    }
  }
  /*
  printf("imaxs %d %d \n",max1,max2);
  printf("max's %lf %lf\n",maxval1, maxval2);
  */


  /* Evaluate deflated value of option price as predicted by Stulz. */
/*
  return(exp(-time*md->option.riskFreeRate)*  \
         stulzMax(maxval1, maxval2, sqrt(md->option.pData.covariance[max1][max1]), \
         sqrt(md->option.pData.covariance[max2][max2]), \
         md->option.pData.covariance[max1][max2], md->option.riskFreeRate, \
         md->option.strikePrice,
         1.0, md->option.pData.dividendRate[max2])); 
*/
  return(exp(-time*md->option.riskFreeRate)* 
     maxOf2Risk(maxval1, maxval2, md->option.strikePrice, 1.0,
     sqrt(md->option.pData.covariance[max1][max1]), 
     sqrt(md->option.pData.covariance[max2][max2]), 
     md->option.riskFreeRate, 
      md->option.pData.dividendRate[max1], 
      md->option.pData.dividendRate[max2],
      md->option.pData.covariance[max1][max2]));
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* The mesh estimate of control3 at time, path */
double innerControl3Estimate(meshData *md, int time, int path){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* At the current time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]);
  maxval1 = md->option.pData.price[time][path][max1];

  for(i=0; i<md->option.pData.underliers;i++){
    if(i != max1 && md->option.pData.price[time][path][i] > maxval2){
      max2 = i;
      maxval2 = md->option.pData.price[time][path][i];
    }
  }
  
  sumWeight = 0;
  estimate = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    sumWeight += md->weights[time][path][toPath];
    if( md->option.pData.price[time+1][toPath][max1] >= md->option.pData.price[time+1][toPath][max2]){
      maxPrice = md->option.pData.price[time+1][toPath][max1];
    }else{
      maxPrice = md->option.pData.price[time+1][toPath][max2];
    }
    if(maxPrice > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate)*(maxPrice - md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath]*deflatedPrice;
    /*    printf("E:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],deflatedPrice); */
  }
  estimate /= sumWeight;
  /*  printf("E:%lf\t%lf\n",sumWeight,exp(-time*md->option.riskFreeRate)*estimate);*/
  return(exp(-time*md->option.riskFreeRate)*estimate);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* The variance of the mesh estimate of control3 at time,path */
double innerControl3Variance(meshData *md, int time, int path){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* At the current time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]);
  maxval1 = md->option.pData.price[time][path][max1];

  for(i=0; i<md->option.pData.underliers;i++){
    if(i != max1 && md->option.pData.price[time][path][i] > maxval2){
      max2 = i;
      maxval2 = md->option.pData.price[time][path][i];
    }
  }
  
  estimate = 0;
  sumWeight =0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if( md->option.pData.price[time+1][toPath][max1] >= md->option.pData.price[time+1][toPath][max2]){
      maxPrice = md->option.pData.price[time+1][toPath][max1];
    }else{
      maxPrice = md->option.pData.price[time+1][toPath][max2];
    }
    if(maxPrice > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate) * (maxPrice-md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath] *deflatedPrice*deflatedPrice;
    sumWeight += md->weights[time][path][toPath];
    /*    printf("V:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],deflatedPrice);*/
  }
  estimate /= sumWeight;
  /*  printf("V:%lf\t%lf\n",sumWeight,estimate);*/
  return(exp(-2*time*md->option.riskFreeRate)*estimate);
}
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* The correlation of the american estimate and control3 */
double innerControl3Correlation(meshData *md, int time, int path){
  int max, toPath;
  double estimate,deflatedPrice,weight,sumWeight;
  double e,p,r, maxPrice;

  int max1, max2, i;
  double maxval1, maxval2;

  /* At the current time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(md->option.pData.underliers, md->option.pData.price[time][path]);
  maxval1 = md->option.pData.price[time][path][max1];

  for(i=0; i<md->option.pData.underliers;i++){
    if(i != max1 && md->option.pData.price[time][path][i] > maxval2){
      max2 = i;
      maxval2 = md->option.pData.price[time][path][i];
    }
  }
  
  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    if( md->option.pData.price[time+1][toPath][max1] >= md->option.pData.price[time+1][toPath][max2]){
      maxPrice = md->option.pData.price[time+1][toPath][max1];
    }else{
      maxPrice = md->option.pData.price[time+1][toPath][max2];
    }

    if(maxPrice > md->option.strikePrice){
      deflatedPrice = exp(-md->option.riskFreeRate) * (maxPrice - md->option.strikePrice);
    }else{
      deflatedPrice = 0;
    }
    estimate += md->weights[time][path][toPath] * exp(-time*md->option.riskFreeRate)*deflatedPrice*md->newValue[time+1][toPath];
    sumWeight += md->weights[time][path][toPath];
    /*    printf("C:%i\t%i\t%lf\t%lf\n",time,path,md->weights[time][path][toPath],exp(-time*md->option.riskFreeRate)*deflatedPrice*md->newValue[time+1][toPath]);*/
  }
  estimate /= sumWeight;
  /*  printf("C:%lf\t%lf\n",sumWeight,estimate);*/
/*
  printf("Correllation = %lf \n",estimate);
*/
  return(estimate);
} 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* Mesh estimate of american price.  Should probably be in mesh.c 
 * Assumes weights to be calculated.  This could be removed from 
 * meshDataFill and made a function in mesh.c
 */
double innerMeshEstimate(meshData *md, int time, int path){
int max, toPath;
  double estimate,sumWeight;
  double w,v;

  estimate = 0;
  sumWeight = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    sumWeight += md->weights[time][path][toPath];
    estimate +=  md->weights[time][path][toPath]*md->newValue[time+1][toPath];
    /*    printf("M:%i\t%i\t%i\t%lf\t%lf\n",time,path,toPath,md->weights[time][path][toPath],md->newValue[time+1][toPath]);*/
  }

  estimate /= sumWeight;
  /*  printf("M:%lf\t%lf\n",sumWeight,estimate);*/
  return(estimate);
} 


/* The controlled value of the american estimate at time, path */
double innerNewEstimate(meshData *md, int time, int path, int whichControl){
  double beta, I1, J1, J1V, J, I1J1, S;
  int toPath, fromPath;

  S = 0;
  for(toPath=0;toPath<md->meshPaths;toPath++){
    S += md->weights[time][path][toPath];
  }
  switch(whichControl){
    case(1):
	  J1   = innerControl1Estimate(md,time,path);    
	  J    = innerControl1Exact(md,time,path);
	  J1V  = innerControl1Variance(md,time,path);
	  I1J1 = innerControl1Correlation(md,time,path);
    break;
    case(2):
	  J1 = innerControl2Estimate(md,time,path); 
	  J = innerControl2Exact(md,time,path); 
	  J1V  = innerControl2Variance(md,time,path); 
	  I1J1 = innerControl2Correlation(md,time,path); 
    break;
    case(3):
	  J1 = innerControl3Estimate(md,time,path); 
	  J = innerControl3Exact(md,time,path); 
	  J1V  = innerControl3Variance(md,time,path); 
	  I1J1 = innerControl3Correlation(md,time,path); 
/*
      printf("Est %lf Exc %lf Var %lf Cor %lf\n",J1,J,J1V,I1J1);
*/
    break;
  }/* end switch */

  I1   = innerMeshEstimate(md,time,path);


  beta = (I1J1 - I1 * J1)/(J1V - J1 * J1);

/*   if( J1/J > 2.0 || J1/J < 0.5 ){  */
/*      if(time != 0 && path < md->meshPaths){  */
/*        for(fromPath=0;fromPath<md->option.pData.paths;fromPath++){ */
/*  	md->weights[time-1][fromPath][path] = 0.0;  */
/*        }  */
/*      } */
/*      printf("N:%i\t%i\t",time,path); */
/*      printf("%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",I1,J1,J,J1V,I1J1,S,beta,-1.0); */
/*  } */ 

/*    printf("N:%i\t%i\t",time,path);  */
/*    printf("%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",I1,J1,J,J1V,I1J1,S,beta,I1 - beta*(J1 - J));  */
  return(I1 - beta*(J1 - J));
}

void innerDataFill(meshData *md, int whichControl){
  int time,path;

  for(time=md->option.pData.times-2;time>0;time--){
    for(path=md->option.pData.paths-1;path>=0;path--){
      md->newValueIfHeld[time][path] = innerNewEstimate(md,time,path,whichControl);
      if(md->newValueIfHeld[time][path] > md->option.payoff[time][path]){
	md->newValue[time][path] = md->newValueIfHeld[time][path];
      }else{
	md->newValue[time][path] = md->option.payoff[time][path];
      }
    }
  }

  md->newValueIfHeld[0][0] = innerNewEstimate(md,0,0,whichControl);
  if(md->newValueIfHeld[0][0] > md->option.payoff[0][0]){
    md->newValue[time][path] = md->newValueIfHeld[0][0];
  }else{
    md->newValue[0][0] = md->option.payoff[0][0];
  }
}

double innerControlledPathEstimate(meshData *md){
  int time, path;
  double value;

  if(md->option.payoff[0][0] >= md->newValueIfHeld[0][0]){
    return(md->option.payoff[0][0]);
  }

  value = 0;
  for(path=md->meshPaths; path < md->option.pData.paths; path++){
    for(time=1; time < md->option.pData.times; time++){
      if(md->option.payoff[time][path] > md->newValueIfHeld[time][path]){
	value += md->option.payoff[time][path];
	break;
      }
    }
  }
  return(value/((double)(md->option.pData.paths-md->meshPaths)));
}
